%% tclp__expand_dcg.pl
%% Predicates to expand DCG grammar clauses

%% Copyright (C) 2004 INRIA Daniel De Rauglaudre, Emmanuel Coquery

%% This library is free software; you can redistribute it and/or
%% modify it under the terms of the GNU Lesser General Public
%% License as published by the Free Software Foundation; either
%% version 2.1 of the License, or (at your option) any later version.

%% This library is distributed in the hope that it will be useful,
%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
%% Lesser General Public License for more details.

%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%                                                                           %%
%%                                   TCLP                                    %%
%%                                                                           %%
%%         Emmanuel Coquery, projet Contraintes, INRIA Rocquencourt          %%
%%                                                                           %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% $Header: /home/yquem/contraintes/coquery/cvs/contraintes/tclpchr/src/tclp/tclp__expand_dcg.pl,v 1.3 2004/02/06 17:12:47 coquery Exp $

:- module(tclp__expand_dcg,[ transform_dcg/6,
                             enable_dcg_tranform/0 ]).

:- use_module(library(lists)).

:- use_module(tclp(tclp__reader)).

transform_dcg(Phrase,Layout,VariableNames,
              NewPhrase,NewLayout,NewVariableNames) :-
        expand_dcg(Phrase,Layout,NewPhrase,NewLayout,NV),
        append(VariableNames,NV,NewVariableNames).

%%expand_dcg((T1 --> TL1),(T2 :- TL2)) :-
%%  VP='TT',                        
%%  T1=..[P|PL],                    
%%  var_int(VP,1,V1),
%%  expand_goal(VP,1,EN,TL1,TL2),
%%  var_int(VP,EN,VN),
%%  append(PL,[V1,VN],PLN),
%%  T2=..[P|PLN], !.

expand_dcg((T1 --> TL1),Layout1,(T2 :- TL2),Layout2,NewVariables) :-

  decompose_location(Layout1,ArrowPos,[Layout_T1,Layout_TL1]),

  VP='$introduced_dcg_var_',                        

  T1=..[P|PL],
  decompose_location(Layout_T1,PPos,Layout_PL),
  
  var_int(VP,1,V1),
  add_var(V1,T1A1,NewVariables),
  
  expand_goal(VP,1,EN,TL1,Layout_TL1,TL2,Layout_TL2,NewVariables),

  var_int(VP,EN,VN),
  add_var(VN,T1A2,NewVariables),

  append(PL,[T1A1,T1A2],PLN),
  append(Layout_PL,[ArrowPos,ArrowPos],Layout_PLN),
  
  T2=..[P|PLN],
  decompose_location(Layout_T2,PPos,Layout_PLN),
  
  decompose_location(Layout2,ArrowPos,[Layout_T2,Layout_TL2]),

  close_list(NewVariables),!.

%% This predicate lists symbols that dont have supplementary arguments
%% when converting dcgs. Currently, the arguments are subject to the
%% transformation, since the relevant symbols are control symbols
no_expand(',',2).
% no_expand(';',2). % still a special case because it is run 'concurrently'
no_expand('->',2).
no_expand(!,0).

%%expand_goal(VP,BN,EN,G1,G2) :- expand_goal2(VP,BN,EN,G1,G2).

%expand_goal(VP,BN,EN,G1,Layout1,G2,Layout2,NewVariables) :-
%        expand_goal2(VP,BN,EN,G1,Layout1,G2,Layout2,NewVariables).

%%expand_goal2(VP,BN,EN,(G1;H1),(G3;H3)) :-
%%  expand_goal1(VP,BN,EN1,G1,G2),
%%  expand_goal2(VP,BN,EN2,H1,H2),
%%  (EN1<EN2 ->
%%     EN=EN2, var_int(VP,EN1,V1), var_int(VP,EN,V), G3=(G2,(V1=V)), H3=H2 ;
%%   EN1>EN2 ->
%%     EN=EN1, var_int(VP,EN,V), var_int(VP,EN2,V2), G3=G2, H3=(H2,(V2=V)) ;
%%   EN=EN1, G3=G2, H3=H2).
%%expand_goal2(VP,BN,EN,G1,G2) :-
%%  expand_goal1(VP,BN,EN,G1,G2).

expand_goal(VP,BN,EN,(G1;H1),Layout1,(G3;H3),Layout3,NewVariables) :-
        !,
        decompose_location(Layout1,SemiColonPos,[Layout_G1,Layout_H1]),
        decompose_location(Layout3,SemiColonPos,[Layout_G3,Layout_H3]),
        
        expand_goal(VP,BN,EN1,G1,Layout_G1,G2,Layout_G2,NewVariables),
        expand_goal(VP,BN,EN2,H1,Layout_H1,H2,Layout_H2,NewVariables),
        (
          EN1<EN2 ->
          EN=EN2,

          var_int(VP,EN1,V1),
          add_var(V1,XV1,NewVariables),

          var_int(VP,EN,V),
          add_var(V,XV,NewVariables),

          G3=(G2,(XV1=XV)),
          Layout_XV1=SemiColonPos,
          Layout_XV=SemiColonPos,
          decompose_location(Layout_EQ,SemiColonPos,[Layout_XV1,Layout_XV]),
          decompose_location(Layout_G3,SemiColonPos,[Layout_G2,Layout_EQ]),
          
          H3=H2,
          Layout_H3=Layout_H2
        ;
          EN1>EN2 ->
          EN=EN1,

          var_int(VP,EN,V),
          add_var(V,XV,NewVariables),
          
          var_int(VP,EN2,V2),
          add_var(V2,XV2,NewVariables),

          G3=G2,
          Layout_G3=Layout_G2,

          H3=(H2,(XV2=XV)),
          Layout_XV2=SemiColonPos,
          Layout_XV=SemiColonPos,
          decompose_location(Layout_EQ,SemiColonPos,[Layout_XV2,Layout_XV]),
          decompose_location(Layout_H3,SemiColonPos,[Layout_H2,Layout_EQ])
        ;
          EN=EN1,

          G3=G2,
          Layout_G3=Layout_G2,

          H3=H2,
          Layout_H3=Layout_H2
        ).
%expand_goal2(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables) :-
%        expand_goal1(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables).

%%expand_goal1(VP,BN,EN,(G1,GL1),(G2,GL2)) :-
%%  expand_goal0(VP,BN,IN,G1,G2),
%%  expand_goal1(VP,IN,EN,GL1,GL2).
%%expand_goal1(VP,BN,EN,G1,G2) :-
%%  expand_goal0(VP,BN,EN,G1,G2).

%%expand_goal(VP,BN,EN,(G1,GL1),Layout1,(G2,GL2),Layout2,NewVariables) :-
%%        !,
%%        decompose_location(Layout1,CommaPos,[Layout_G1,Layout_GL1]),
%%        decompose_location(Layout2,CommaPos,[Layout_G2,Layout_GL2]),
%%        expand_goal(VP,BN,IN,G1,Layout_G1,G2,Layout_G2,NewVariables),
%%        expand_goal(VP,IN,EN,GL1,Layout_GL1,GL2,Layout_GL2,NewVariables).

expand_goal(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables) :-
        functor(G1,P,N), no_expand(P,N),!,
        decompose_location(Layout_G1,PPos,Layout_GL1),
        G1=..[P|GL1],
        expand_goals(VP,BN,EN,GL1,Layout_GL1,GL2,Layout_GL2,NewVariables),
        G2=..[P|GL2],
        decompose_location(Layout_G2,PPos,Layout_GL2).

%%expand_goal1(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables) :-
%        expand_goal0(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables).

%%expand_goal0(_,BN,BN,{G1},G1).
%%expand_goal0(_,BN,BN,!,!).
%%expand_goal0(VP,BN,EN,G1,(VB=VV)) :-
%%  is_list(G1),
%%  EN is BN+1,
%%  var_int(VP,BN,VB),
%%  var_int(VP,EN,VE),
%%  append(G1,VE,VV).
%%expand_goal0(VP,BN,EN,G1,G2) :-
%%  G1=..[P|PL1],
%%  EN is BN+1,
%%  var_int(VP,BN,VB),
%%  var_int(VP,EN,VE),
%%  append(PL1,[VB,VE],PL2),
%%  G2=..[P|PL2].

expand_goal(_,BN,BN,{G1},Layout1,G1,Layout2,_) :-
        !,
        decompose_location(Layout1,_,[Layout2]).
expand_goal(VP,BN,EN,M:G1,Layout1,M:G2,Layout2,NewVariables) :-
        !,
        decompose_location(Layout1,ColonPos,[Layout_M,Layout_G1]),
        expand_goal(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables),
        decompose_location(Layout2,ColonPos,[Layout_M,Layout_G2]).
expand_goal(VP,BN,EN,G1,Layout_G1,(XVB=VV),Layout_EQ,NewVariables) :-
        is_list(G1), !,
        EN is BN+1,
        decompose_location(Layout_G1,Pos,_),
        
        var_int(VP,BN,VB),
        add_var(VB,XVB,NewVariables),
        
        var_int(VP,EN,VE),
        add_var(VE,XVE,NewVariables),
        
        append(G1,XVE,VV),      % here we replace in VV the end of the list
                                % G1 by the variable XVE, which means that
                                % the layout is the same
        
        decompose_location(Layout_EQ,Pos,[Pos,Layout_G1]).
expand_goal(VP,BN,EN,G1,Layout_G1,G2,Layout_G2,NewVariables) :-
        EN is BN+1,
        G1=..[P|PL1],
        decompose_location(Layout_G1,PPos,Layout_PL1),

        
        var_int(VP,BN,VB),
        add_var(VB,XVB,NewVariables),
        
        var_int(VP,EN,VE),
        add_var(VE,XVE,NewVariables),
        
        append(PL1,[XVB,XVE],PL2),
        append(Layout_PL1,[PPos,PPos],Layout_PL2),
        
        G2=..[P|PL2],
        decompose_location(Layout_G2,PPos,Layout_PL2).

expand_goals(_,BN,BN,[],[],[],[],_).
expand_goals(VP,BN,EN,
             [G1|GL1],[Layout_G1|Layouts_GL1],
             [G2|GL2],[Layout_G2|Layouts_GL2],
             NewVariables) :-
        expand_goal(VP,BN,EN1,G1,Layout_G1,G2,Layout_G2,NewVariables),
        expand_goals(VP,EN1,EN,GL1,Layouts_GL1,GL2,Layouts_GL2,NewVariables).

%% Builds a variable name V with the prefix P and the number N
var_int(P,N,V) :- number_atom(N,B), atom_concat(P,B,V).

%% A is the atom corresponding to the decimal representation of the number N
number_atom(N,A) :-
        nonvar(N),!,
        number_codes(N,CodeList), atom_codes(A,CodeList).
number_atom(N,A) :-
        atom_codes(A,CodeList), number_codes(N,CodeList).

%% Handles list of Name=Variable with a variable as the end of the list
%% If Name is in the list then unifies Var with the corresponding variable
%% in the list, otherwise, adds the pair at the end of the list
add_var(Name,Var,[Name=Var|_]) :- !.
add_var(Name,Var,[_|L]) :- add_var(Name,Var,L).

%% If the end of the list is a variable, the instanciate it with []
close_list(X) :- var(X), !, X=[].
close_list([_|L]) :- close_list(L).

%% Decomposes a layout the line of the head symbol + a list of the layout
%% of its arguments
decompose_location(Loc,Pos,Args) :- var(Loc),!,
        (
          Args = []
        ->
          Pos=Loc
        ;
          Loc=[Pos|Args]
        ).
decompose_location([X|Args],X,Args) :- !.
decompose_location(X,X,[]).

%% Adds the dcg expansion to phrase handlers
enable_dcg_tranform :- add_phrase_transformer(tclp__expand_dcg:transform_dcg).

%% Changement dans expand_goal a cause des parenthses, qui n'apparaissent
%% pas dans le terme lu